import sys
from pathlib import Path

path = Path(__file__).parent
sys.path.insert(0, str((path).resolve()))
sys.path.insert(0, str((path / "../../src").resolve()))

from clig import Command, Arg, data


def test_subparsers_with_same_parameters_all_kw():
    def maincmd(foo: str | None = None):
        assert foo == "yoco"

    def subcmd(foo: str | None = None):
        assert foo == "rocky"

    def subsubcmd(foo: str | None = None):
        assert foo == "sand"

    (
        Command(maincmd)
        .new_subcommand(subcmd)
        .end_subcommand(subsubcmd)
        .run("--foo yoco subcmd --foo rocky subsubcmd --foo sand".split())
    )



from pathlib import Path
import os

os.chdir(Path(__file__).parent)

os.system("jupyter nbconvert notebooks/userguide.ipynb --to markdown --output userguide.md")

with open("notebooks/userguide.md", "r", encoding="utf-8") as file:
    text = file.read()

text = text.replace("```python\n! python", "```\n> python")


lines = text.split("\n")

on_shell_snippet = False
on_python_snippet = False
on_python_snippet_output = False
end_of_python_snippet_output = False
on_python_snippet_decorator = False
snippet_started = False
for i, line in enumerate(lines):
    if line.strip().startswith("Couldn't find"):
        lines[i] = "<must_remove>"
    if line.startswith("%%python"):
        on_python_snippet = False
        lines[i] = "<must_remove>"
    if line.startswith("```bash") or line.startswith("> python"):
        on_shell_snippet = True
        continue
    if line.startswith("```python"):
        on_python_snippet = True
        continue
    if on_python_snippet:
        if line.startswith("```"):  # end of snippet containing python code. Need to continue for output
            lines[i] = "<must_remove>"  # remove this single line
            on_python_snippet = False  # end of snippet containing python code
            on_python_snippet_output = True  # the next is the output
            continue
        if line.startswith(" "):
            lines[i] = "... " + lines[i]
            continue
        if line.startswith("@"):
            lines[i] = ">>> " + lines[i]
            on_python_snippet_decorator = True
            continue
        if on_python_snippet_decorator:
            lines[i] = "... " + lines[i]
            on_python_snippet_decorator = False
            continue
        if len(line) == 0:
            lines[i] = "... " + lines[i]
            continue
        if line.startswith("```"):
            on_python_snippet = False
            continue
        lines[i] = ">>> " + lines[i]
    if on_python_snippet_output:
        if not line.startswith("    "):  # end or start of output snippet
            if end_of_python_snippet_output:  # end of output snippet
                lines[i] = "```"  # finalize output snippet
                on_python_snippet_output = False
                end_of_python_snippet_output = False
                continue
            else:
                lines[i] = "<must_remove>"  # remove this single line
                end_of_python_snippet_output = True  # the next will be the end of output snippet
                continue
        else:
            lines[i] = lines[i].strip()
            if len(lines[i]) == 0:
                lines[i] = "<must_remove>"
    if on_shell_snippet:
        if line.startswith("```"):  # end of notebook snippet containing single line with cli command
            lines[i] = "<must_remove>"  # remove this single line
            continue
        if not line.startswith("    "):  # end or start of output snippet
            if snippet_started:  # end of output snippet, because it already started
                lines[i] = "```"  # finalize output snippet
                on_shell_snippet = False
                snippet_started = False
                continue
            else:
                snippet_started = True  # start of output snippet
                continue


text = "\n".join([line for line in lines if line != "<must_remove>"]).replace("... \n...", "...")
with open("notebooks/userguide.md", "w", encoding="utf-8") as file:
    file.write(text)

# with open("../../../src/clig/__init__.pyi", "w", encoding="utf-8") as file:
#     file.write(
#         '"""\n'
#         + text.replace('"""', '\\"\\"\\"')
#         .replace("```\n\n```", "```\n`\n```")
#         .replace("```\n\n\n```", "```\n`\n```")
#         + '"""\n'
#     )



import inspect
import os
import sys

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__) + "/../src"))
from clig import clig

from typing import Literal, Sequence


def test_get_data_from_argtype_simple_type():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(ascii)
    assert action == "store"
    assert nargs == None
    assert argtype == ascii
    assert choices is None


def test_get_data_from_argtype_Literal():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(Literal["option1", "option2"])
    assert action == "store"
    assert nargs == None
    assert argtype == None
    assert choices is not None
    assert set(choices) == set(["option1", "option2"])


def test_get_data_from_argtype_List():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(list[str])
    assert action == "store"
    assert nargs == "*"
    assert argtype == str
    assert choices is None


def test_get_data_from_argtype_Sequence():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(Sequence[int])
    assert action == "store"
    assert nargs == "*"
    assert argtype == int
    assert choices is None


def test_get_data_from_argtype_Tuple():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(tuple[int, int, int])
    assert action == "store"
    assert nargs == 3
    assert argtype == int
    assert choices is None


def test_get_data_from_argtype_Tuple_Ellipsis():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(tuple[float, ...])
    assert action == "store"
    assert nargs == "*"
    assert argtype == float
    assert choices is None


def test_get_data_from_argtype_Bool():
    action, nargs, argtype, choices = clig._get_data_from_typeannotation(bool)
    assert action == "store_true"
    assert nargs == None
    assert argtype == None
    assert choices is None



import sys
from pathlib import Path

path = Path(__file__).parent
sys.path.insert(0, str((path).resolve()))
sys.path.insert(0, str((path / "../src").resolve()))

from clig import Command
import functions as fun


def test_posWithType_kwWithType_kwBoolWithType_cligDocMultiline():
    cmd = Command(fun.ptc_kti_ktb_cligDocMutiline)
    cmd._add_parsers()



"""
Legend:

- p = positional
- k = keyword
- n = no type annotation
- t = with type annotation
- m = with metadata in type annotation
- o = optional (type union with None default to None)

- i = integer
- f = float
- b = boolean
- c = string (chars)
- u = tuple
- l = list
- s = sequence
- a = Literal
"""

# cSpell:disable
import sys
from pathlib import Path

sys.path.insert(0, str((Path(__file__).parent / "/../src").resolve()))
from clig import Arg, data


def pn_noDoc(a):
    pass


def pn_knc_noDoc(first, second="test"):
    pass


def pn_pt_kti_noDoc(a, b: float, c: int = 123):
    pass


def ptcm_ptcm_ktb(
    a: Arg[str, data("-f", "--first", help="The first argument")],
    b: Arg[int, data(action="store_const", const=123)],
    c: bool = True,
):
    pass


def ptc_kti_ktb_cligDocMutiline(a: str, b: int = 123, c: bool = True):
    """Reprehenderit unde commodi doloremque rerum ducimus quam accusantium.

    Qui quidem quo eligendi officia ea quod ab tempore esse. Sapiente quasi est sint. Molestias et
    laudantium quidem laudantium animi voluptate asperiores illum. Adipisci tempora nesciunt dolores
    tempore consequatur amet. Aut ipsa ex.

    Parameters
    ----------
    - `a` (`str`):
        Dicta et optio dicta.

    - `b` (`int`, optional): Defaults to `123`.
        Dolorum voluptate voluptas nisi.

    - `c` (`bool`, optional): Defaults to `True`.
        Asperiores quisquam odit voluptates et eos incidunt. Maiores minima provident doloremque aut
        dolorem. Minus natus ab voluptatum totam in. Natus consectetur modi similique rerum excepturi
        delectus aut.

    """
    pass


def descr():
    """A foo that bars"""
    pass


def descrEpilog():
    """Aliquam alias quia earum.

    Corporis ullam nam ut dolores sed. Nemo ea deserunt facere numquam velit aut. Architecto provident
    consequatur ratione est quas qui dolor ratione. Laudantium fugit at.

    Ullam et temporibus eum. Sit voluptatem tempora totam dolores. Pariatur accusamus voluptate totam.
    Fugit rerum nemo reiciendis veritatis modi sit distinctio ratione id.

    Voluptates tenetur quos qui exercitationem laudantium aliquid. Neque qui eum qui. Qui tenetur facilis
    non voluptatem ut corporis harum fugiat.
    """
    pass


def pti_ptc_ptf_ktb_ktlo_numpyEpilog(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Distinctio et ratione sequi hic.

    Blanditiis velit consequatur omnis odit magnam quo dignissimos. Qui ex et illo. Et
    necessitatibus ea placeat consectetur itaque dolore fugiat quo autem. Ut accusamus incidunt
    repellat minima est soluta ut est. Id aut enim ad. Quia qui sint ex eos eveniet eveniet
    earum unde.

    Non rerum aut consectetur ut ducimus ut similique ut illum. Aut qui et distinctio. Nihil id
    sit incidunt minus omnis. Quo unde inventore fuga. Quasi ea ea dolores quam.

    Esse temporibus voluptas nulla. Odio voluptas nisi quae cupiditate consequatur cumque ut ex
    dolorem. Est doloremque quis nostrum voluptates doloremque quia. Ex sunt dolores consectetur
    veritatis maxime suscipit. Fugit dolorem facilis quasi.

    Parameters
    ----------
    a : int
        Neque ut qui non nulla odit esse accusantium aut suscipit.
    b : str
        Nihil dolores autem autem nulla sit nihil molestiae vero est.
    c : float
        Animi magnam ut sapiente maiores.
    d : bool, optional
        Consequatur provident neque optio consequatur., by default True
    e : list[str] | None, optional
        Corrupti molestiae in aspernatur., by default None

    Returns
    -------
    tuple[str, ...]
        Numquam maiores atque doloribus.
    """
    pass


def pti_ptc_ptf_ktb_ktlo_sphinxEpilog(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Qui accusantium harum debitis et.

    Est nam quia voluptatem vero architecto laborum. Accusantium delectus et aut repudiandae
    voluptatibus qui iure ut debitis. Voluptatibus ut enim consequatur iusto eaque dolor.

    :param a: Atque pariatur excepturi sed dolorem sint impedit molestiae.
    :type a: int
    :param b: magni optio voluptatibus
    :type b: str
    :param c: Sit eligendi consequatur recusandae doloribus enim amet.
    :type c: float
    :param d: Soluta dolorum amet et., defaults to True
    :type d: bool, optional
    :param e: Fugiat provident amet iste natus ab voluptas., defaults to None
    :type e: list[str] | None, optional
    :return: Beatae perspiciatis ut in incidunt vitae.
    :rtype: tuple[str, ...]
    """
    pass


def pti_ptc_ptf_ktb_ktlo_googleEpilog(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Voluptatum dolorem quis dolorum voluptas atque non temporibus.

    Maiores occaecati quam asperiores non sunt est dolor laborum est. Eius corporis nobis
    accusamus rerum et et et. Ducimus tempore voluptas qui aut consectetur saepe quos cum
    delectus. Tempora adipisci odit qui. Optio eum magni non. Placeat repudiandae quasi nostrum
    mollitia sunt neque fuga id possimus.

    Args:
        a (int): Qui eum eius nihil voluptas quia aut numquam.
        b (str): Quasi voluptates dicta cumque similique qui dolorem architecto.
        c (float): Sint harum et omnis nobis numquam quos omnis.
        d (bool, optional): Ex voluptas animi.. Defaults to True.
        e (list[str] | None, optional): In vero ut nisi officia ut.. Defaults to None.

    Returns:
        tuple[str, ...]: Pariatur aut asperiores aut omnis maxime ratione nemo ut.
    """
    pass


def pti_ptc_ptf_ktb_ktlo_cligEpilog(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Fugit voluptatibus enim odit velit facilis.

    Neque dolores expedita repellat in perspiciatis dolorem aliquid et. Commodi fugit minima
    laudantium beatae et ut. Id possimus soluta magnam quisquam laboriosam impedit.

    Ad quaerat ut culpa aut iure id quia. Ut aut alias adipisci quia. Veritatis ratione
    dignissimos laborum. Molestiae molestias id earum.

    Nesciunt quas corrupti tenetur officiis occaecati asperiores eaque. Qui voluptas ut ea dolor
    et harum beatae quos. Est tenetur ut ipsum. Eveniet rem beatae error eum voluptatem tempora
    velit in. Ea doloribus similique.

    Parameters
    ----------
    - `a` (`int`):
        Quidem natus sunt molestiae et reprehenderit voluptas optio.

    - `b` (`str`):
        Unde rerum aut a et assumenda fugit dolorem eligendi corrupti.

    - `c` (`float`):
        Dolorum officiis totam aspernatur fuga voluptas similique.

    - `d` (`bool`, optional): Defaults to `True`.
        Ducimus sunt eum in vel voluptatibus aut facere perspiciatis.

    - `e` (`list[str] | None`, optional): Defaults to `None`.
        Sit et consequatur a asperiores sequi sint dolores id ipsam.

    Returns
    -------
    `tuple[str, ...]`:
        illo odit ut
    """
    pass


def ptc_ptb_cligEpilog(name: str, flag: bool):
    """Ullam non nulla et dolore.

    Tempora accusantium aliquid unde magnam hic voluptate omnis consequatur. Officia consequatur facere
    voluptatibus quo modi ea aliquid fugiat. Autem officia ipsa in quaerat itaque. Numquam omnis nihil laborum
    quae hic voluptas ut. Voluptatem ut asperiores aut omnis nesciunt ducimus dolorum. Quis quos officia
    voluptatibus quod assumenda laborum.

    Parameters
    ----------
    - `name` (`str`):
        Sequi deserunt est quia qui.

    - `flag` (`bool`):
        Labore eius et voluptatem quos et consequatur dolores.

    """
    pass


def pti_ptc_ptf_ktb_ktlo_numpyEpilogMultiline(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Voluptatibus eos ipsa ex debitis voluptatem dignissimos.

    Qui deserunt sequi aut illo. Minima modi illo sit occaecati. Ducimus illum et. Deleniti repellendus
    cum quasi ut et natus dolorem. Aliquam aut aperiam qui.

    Sint maiores dolorum. Nobis quo distinctio consequuntur. Recusandae fuga qui perspiciatis quisquam.
    Nostrum asperiores neque nisi. Enim voluptatem eum illo.

    Est labore illum voluptatibus at ut. Deleniti ut aut ut aperiam consequatur aut. Reprehenderit
    voluptatem est voluptates temporibus et voluptate accusamus dolores. Placeat nihil dignissimos sed
    sequi sequi.

    Parameters
    ----------
    a : int
        Fuga nemo provident vero odio qui sint et aut veritatis. Facere necessitatibus ut. Voluptatem
        natus natus veritatis earum. Reprehenderit voluptate dolorem dolores consequuntur magnam impedit
        eius. Est ut nisi aut accusamus.
    b : str
        Culpa asperiores incidunt molestias aliquam soluta voluptas excepturi nulla.
    c : float
        Non vitae qui non magni harum eum maiores qui. Dicta sunt voluptatem voluptate. At quibusdam
        aliquam autem et perspiciatis et assumenda. Perferendis qui velit quam numquam iste.
    d : bool, optional
        Aut ipsam aut velit impedit. Quidem expedita aliquid sed officia in ex et. Nihil rem adipisci ut
        perferendis iure., by default True
    e : list[str] | None, optional
        Ratione consequatur molestiae quia deserunt quo. Non cupiditate sunt commodi vero labore
        doloremque voluptatem officiis est. Iusto voluptate reiciendis iusto in. Occaecati quia soluta
        minus perspiciatis alias illum iste aperiam et. Autem accusamus unde omnis est cum ducimus. Iure
        adipisci id omnis quis placeat impedit rerum ab aspernatur.

        Praesentium id rerum quod provident odit dolores adipisci veniam natus. Porro repellat aliquid
        quibusdam recusandae hic voluptas accusantium voluptatem voluptatem. Laboriosam similique nobis
        aut iusto et ab minima cum.

        Voluptas molestiae mollitia autem distinctio magnam dolorem molestiae aliquid. Neque provident
        impedit et. Quod quibusdam nulla cupiditate. Praesentium neque vel ea velit consequatur quis
        voluptate iste. Quae veniam sequi et nihil qui vel voluptatem maxime. Laborum corrupti dolores
        voluptate placeat fugit non nobis., by default None
    """
    pass


def pti_ptc_ptf_ktb_ktlo_sphinxEpilogMultiline(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """Est sit minus quasi soluta unde vero deleniti eligendi.

    Quia aspernatur doloribus id doloribus sunt ratione et voluptatum. Eligendi numquam sed. Voluptas
    consequuntur quibusdam debitis quia unde doloribus ducimus sunt. Et provident assumenda hic eum sint
    quia ipsum sit sed.

    :param a: Velit ratione harum in quia laborum ut est quis.
    :type a: int
    :param b: Adipisci voluptates aut fugiat qui nam non. Eveniet molestiae voluptas explicabo fuga.
        Beatae ex sed nostrum incidunt.
    :type b: str
    :param c: Non non voluptatum ipsum sit maiores et eum adipisci. Autem sit possimus et similique atque.
        Nihil tempore et excepturi.

        Nisi magnam et. Illum minus ea enim eligendi doloremque consequatur odit est officiis. Dolorem
        dolores repellat esse vero quae. Laboriosam ab qui quo eveniet quia ex et. Aut facilis molestias
        qui. Dolorum sit magni repellat iusto aut vel.

        Laborum dolores illum modi. Id et qui nisi harum aperiam doloribus. Quod quibusdam dolorum iusto.
    :type c: float
    :param d: Et magni harum adipisci accusantium aut et ipsum impedit. Sit modi voluptatem. Esse quis aut
        ex. Dicta quam rem repellendus accusantium aut molestias praesentium fugiat corporis. Assumenda
        eum natus voluptatem alias dolorem vitae dolor repudiandae inventore. Et deleniti repellendus
        quo., defaults to True
    :type d: bool, optional
    :param e: Corporis est rerum. Aspernatur dolor porro a culpa omnis. Repudiandae totam necessitatibus
        quibusdam ipsum numquam eveniet dolor quasi. Dolores dolorem voluptate aut. Deleniti officia qui
        molestiae. Quo deserunt nulla aut sit sunt quam nostrum odit et., defaults to None
    :type e: list[str] | None, optional
    """
    pass


def pti_ptc_ptf_ktb_ktlo_googleEpilogMultiline(
    a: int, b: str, c: float, d: bool = True, e: list[str] | None = None
) -> None:
    """nesciunt beatae asperiores

    Et perferendis quia et sit maxime. Accusantium vel sint quam perspiciatis minus explicabo. Incidunt
    iste error autem impedit deserunt tempore quo aut odit.

    Args:
        a (int): Vel similique placeat. Nam enim perspiciatis qui earum voluptas quis. Perspiciatis ut
            vitae. Aspernatur ab ratione libero ex hic consequatur. Nam cupiditate earum. Nihil ea
            exercitationem ut.
        b (str): Molestiae velit et expedita autem quam. Omnis dolorem placeat est. Quidem illum eveniet
            enim exercitationem aut qui dolore est et. Rerum est iste laudantium qui praesentium et. Et
            deserunt voluptates harum voluptas voluptates iste saepe consequatur.
        c (float): Quae minima eligendi veniam aperiam libero temporibus quia qui atque. Velit ea aut vel
            quibusdam commodi id laboriosam inventore aliquam. Nam nisi itaque et sed dolor praesentium
            molestiae quisquam cupiditate. Voluptatem mollitia dolorem est deleniti repellat cum
            voluptatem voluptas sit.

            Sit animi dolore neque libero voluptatibus. Illum voluptatum ullam distinctio quisquam sequi
            delectus quia similique sit. Id enim vel eius iure rerum veritatis eos rem et. Nemo est
            assumenda aut et quo. Et soluta corrupti amet perferendis maxime.

            Placeat aut consequatur vel quo impedit doloribus et in libero. Voluptas ducimus suscipit.
            Assumenda alias est sed asperiores similique id consequuntur. Voluptas rerum placeat
            perferendis possimus ratione at. Ea ut aut id explicabo voluptas.
        d (bool, optional): Error ut architecto fugit natus qui tempora vitae. A sed sequi reprehenderit
            quia autem voluptatem enim. Numquam cum minus cum eos est. Illo voluptas ducimus minus ipsam
            quae dolores quam quo. Quod qui sed incidunt rerum sed. Incidunt repellendus est est labore
            laudantium quia voluptas ipsum.. Defaults to True.
        e (list[str] | None, optional): Explicabo tenetur beatae consequuntur atque aut omnis et. Eveniet
            ipsum repellat voluptatibus sit.

            Placeat eum veritatis praesentium voluptates quia beatae
            repellendus suscipit. Sint neque deserunt quis. Incidunt quibusdam voluptatem animi voluptas
            in. Voluptas dolor aut quisquam.. Defaults to None.
    """
    pass


def ptc_kti_ktf_clig(a: str, b: int = 666, c: float = 3.14):
    """Incidunt odio dolorum quia blanditiis quis doloremque unde. Sapiente nemo illum facere dolores sunt
    veniam minus. Id doloremque dicta accusamus fuga ut qui nesciunt.

    Parameters
    ----------
    - `a` (`str`):
        Pariatur quis voluptates nemo eum occaecati.

    - `b` (`int`, optional): Defaults to `666`.
        Odio nostrum cupiditate quod debitis quaerat tempore eveniet excepturi.

    - `c` (`float`, optional): Defaults to `3.14`.
        Aut illo necessitatibus optio.
    """
    pass


def pn_pn_knb_kni_numpyEpilogNoType(a, b, c=False, d=123):
    """Odio est rerum iure porro rerum voluptatum libero magnam.

    In vitae ut distinctio optio corrupti cumque rerum quasi aut.

    Parameters
    ----------
    a
        hic omnis sint
    b
        Ut rem quis delectus.
    c
        Et tenetur modi ea., by default False
    d
        recusandae autem vero, by default 123
    """
    pass


def pn_pn_knb_kni_googleEpilogNoType(a, b, c=False, d=123):
    """Odio est rerum iure porro rerum voluptatum libero magnam.

    In vitae ut distinctio optio corrupti cumque rerum quasi aut.

    Args:
        a: Quasi veniam facere et.
        b: Quis ex modi vel sed ea dolorum magnam. Ut veniam veniam minus. Laboriosam voluptatem quod et. Et
            eaque sint quasi libero mollitia.
        c: architecto non voluptas. Defaults to False.
        d: Omnis laboriosam aut saepe nobis consequatur nihil eos accusantium.. Defaults to 123.
    """
    pass


def pn_pn_knb_kni_googleNoEpilogNoType(a, b, c=False, d=123):
    """Odio est rerum iure porro rerum voluptatum libero magnam.

    Args:
        a: Quasi veniam facere et.
        b: Quis ex modi vel sed ea dolorum magnam. Ut veniam veniam minus. Laboriosam voluptatem quod et. Et
            eaque sint quasi libero mollitia.
        c: architecto non voluptas. Defaults to False.
        d: Omnis laboriosam aut saepe nobis consequatur nihil eos accusantium.. Defaults to 123.
    """
    pass


def pn_pn_knb_kni_sphinxEpilogNoType(a, b, c=False, d=123):
    """Odio est rerum iure porro rerum voluptatum libero magnam.

    In vitae ut distinctio optio corrupti cumque rerum quasi aut.

    :param a: aperiam enim voluptate
    :param b: Totam voluptas porro est sint.
    :param c: Iusto impedit numquam ut., defaults to False
    :param d: Corporis quis fugit eveniet rerum., defaults to 123
    """
    pass



##############################################################################################################
# %%          IMPORTS
##############################################################################################################

from __future__ import annotations

import inspect
import re
import sys
from argparse import ArgumentParser, FileType, HelpFormatter, Action, BooleanOptionalAction, Namespace
from argparse import HelpFormatter, RawTextHelpFormatter, _SubParsersAction  # [_ArgumentParserT]
from dataclasses import KW_ONLY, Field, dataclass, field
from inspect import Parameter
from inspect import _ParameterKind
from types import MappingProxyType, UnionType
from collections import OrderedDict
from collections.abc import Sequence
from typing import get_args, get_origin, Union, Annotated
from typing import Any, Callable, Iterable, Literal, Mapping, Self, TypedDict, Unpack, overload
from enum import Enum, StrEnum

Kind = _ParameterKind
Arg = Annotated

EMPTY = Parameter.empty

##############################################################################################################
# %%          DOCSTRINGS TEMPLATES
##############################################################################################################

DESCRIPTION_DOCSTRING = """{{description}}"""

DESCRIPTION_EPILOG_DOCSTRING = """
    {{description}}    

    {{epilog}}
"""

NUMPY_DOCSTRING_WITH_EPILOG = """
    {{description}}

    {{epilog}}

    Parameters
    ----------
    {{parameter_name}} : {{parameter_type}}
        {{parameter_description}}
"""

NUMPY_DOCSTRING_WITH_EPILOG_NOTYPES = """
    {{description}}

    {{epilog}}

    Parameters
    ----------
    {{parameter_name}}
        {{parameter_description}}
"""

SPHINX_DOCSTRING_WITH_EPILOG = """
{{description}}

{{epilog}}

:param {{parameter_name}}: {{parameter_description}}
:type {{parameter_name}}: {{parameter_type}}
"""

SPHINX_DOCSTRING_WITH_EPILOG_NOTYPES = """
{{description}}

{{epilog}}

:param {{parameter_name}}: {{parameter_description}}
"""

GOOGLE_DOCSTRING_WITH_EPILOG = """
{{description}}

{{epilog}}

Args:
    {{parameter_name}} ({{parameter_type}}): {{parameter_description}}
"""

GOOGLE_DOCSTRING_WITH_EPILOG_NOTYPES = """
{{description}}

{{epilog}}

Args:
    {{parameter_name}}: {{parameter_description}}
"""

GOOGLE_DOCSTRING_NOTYPES = """
{{description}}

Args:
    {{parameter_name}}: {{parameter_description}}
"""

CLIG_DOCSTRING_WITH_EPILOG = """
{{description}}

{{epilog}}

Parameters
----------
- `{{parameter_name}}` {{parameter_type}}
    {{parameter_description}}
"""


NUMPY_DOCSTRING = """
    {{description}}

    Parameters
    ----------
    {{parameter_name}} : {{parameter_type}}
        {{parameter_description}}
"""

SPHINX_DOCSTRING = """
{{description}}

:param {{parameter_name}}: {{parameter_description}}
:type {{parameter_name}}: {{parameter_type}}
"""

GOOGLE_DOCSTRING = """
{{description}}

Args:
    {{parameter_name}} ({{parameter_type}}): {{parameter_description}}
"""

CLIG_DOCSTRING = """
{{description}}

Parameters
----------
- `{{parameter_name}}` {{parameter_type}}
    {{parameter_description}}
"""

# TODO: add 'no type' variants

DOCSTRING_TEMPLATES = [
    DESCRIPTION_DOCSTRING,
    DESCRIPTION_EPILOG_DOCSTRING,
    NUMPY_DOCSTRING_WITH_EPILOG,
    SPHINX_DOCSTRING_WITH_EPILOG,
    GOOGLE_DOCSTRING_WITH_EPILOG,
    CLIG_DOCSTRING_WITH_EPILOG,
    NUMPY_DOCSTRING_WITH_EPILOG_NOTYPES,
    SPHINX_DOCSTRING_WITH_EPILOG_NOTYPES,
    GOOGLE_DOCSTRING_WITH_EPILOG_NOTYPES,
    GOOGLE_DOCSTRING_NOTYPES,
    NUMPY_DOCSTRING,
    SPHINX_DOCSTRING,
    GOOGLE_DOCSTRING,
    CLIG_DOCSTRING,
]

SUBPARSERS_DEST = "subcommand_"


class DocStr(StrEnum):
    DESCRIPTION_DOCSTRING = DESCRIPTION_DOCSTRING
    DESCRIPTION_EPILOG_DOCSTRING = DESCRIPTION_EPILOG_DOCSTRING
    NUMPY_DOCSTRING_WITH_EPILOG = NUMPY_DOCSTRING_WITH_EPILOG
    NUMPY_DOCSTRING_WITH_EPILOG_NOTYPES = NUMPY_DOCSTRING_WITH_EPILOG_NOTYPES
    SPHINX_DOCSTRING_WITH_EPILOG = SPHINX_DOCSTRING_WITH_EPILOG
    SPHINX_DOCSTRING_WITH_EPILOG_NOTYPES = SPHINX_DOCSTRING_WITH_EPILOG_NOTYPES
    GOOGLE_DOCSTRING_WITH_EPILOG = GOOGLE_DOCSTRING_WITH_EPILOG
    GOOGLE_DOCSTRING_WITH_EPILOG_NOTYPES = GOOGLE_DOCSTRING_WITH_EPILOG_NOTYPES
    CLIG_DOCSTRING_WITH_EPILOG = CLIG_DOCSTRING_WITH_EPILOG
    NUMPY_DOCSTRING = NUMPY_DOCSTRING
    SPHINX_DOCSTRING = SPHINX_DOCSTRING
    GOOGLE_DOCSTRING = GOOGLE_DOCSTRING
    CLIG_DOCSTRING = CLIG_DOCSTRING
    GOOGLE_DOCSTRING_NOTYPES = GOOGLE_DOCSTRING_NOTYPES


##############################################################################################################
# %%          MAIN CLASS
##############################################################################################################


@dataclass
class Command:
    func: Callable[..., Any] | None = None
    # Arguments for `ArgumentParser` object, see: https://docs.python.org/3/library/argparse.html#argumentparser-objects
    prog: str | None = None
    usage: str | None = None
    description: str | None = None
    epilog: str | None = None
    parents: Sequence[ArgumentParser] = field(default_factory=list)
    formatter_class: type[HelpFormatter] = RawTextHelpFormatter
    prefix_chars: str = "-"
    fromfile_prefix_chars: str | None = None
    argument_default: Any = None
    conflict_handler: str = "error"
    add_help: bool = True
    allow_abbrev: bool = True
    exit_on_error: bool = True
    # Arguments for Sub-commands, see: https://docs.python.org/3/library/argparse.html#sub-commands
    _: KW_ONLY
    subcommands_title: str = "subcommands"
    subcommands_description: str | None = None
    subcommands_prog: str | None = None
    subcommands_required: bool = False
    subcommands_help: str | None = None
    subcommands_metavar: str | None = None
    # Extra arguments of this library
    docstring_template: str | DocStr | None = None
    default_bool: bool = False
    name: str | None = None
    help: str | None = None
    aliases: Sequence[str] = field(init=False, default_factory=list)
    parent: Command | None = field(init=False, default=None)
    parser: ArgumentParser | None = field(init=False, default=None)
    # TODO: `make_flags` option general
    # TODO: `make_shorts` option
    # TODO: `make_longs` option
    # TODO: set `func` before init

    def __post_init__(self):

        self.parameters: Mapping[str, Parameter] = {}
        """A dict with `name: Parameter`, where `Parameter` comes from stdlib `inspect`
        ref: https://docs.python.org/3/library/inspect.html#inspect.Parameter"""

        if self.func:
            self.name = self.name or self.func.__name__
            self.parameters = inspect.signature(self.func).parameters

        self.docstring_data: _DocstringData | None = self._get_data_from_docstring()
        self.argument_data: list[_ArgumentData] = self.__generate_argument_data_list()
        if self.docstring_data:
            self.description = self.description or self.docstring_data.description
            self.epilog = self.epilog or self.docstring_data.epilog

        self.sub_commands: OrderedDict[str, Command] = OrderedDict()
        self.sub_commands_group: _SubParsersAction | None = None
        self.longstartflags: str = f"{self.prefix_chars}" * 2

    ##########################################################################################################
    # %:          PUBLIC METHODS
    ##########################################################################################################

    @overload
    def subcommand[**P, T](self, func: Callable[P, T]) -> Callable[P, T]: ...

    @overload
    def subcommand[**P, T](self, **kwargs) -> Callable[[Callable[P, T]], Callable[P, T]]: ...

    def subcommand[**P, T](
        self,
        func: Callable[P, T] | None = None,
        **kwargs,
    ) -> Callable[P, T] | Callable[[Callable[P, T]], Callable[P, T]]:  # fmt: skip
        """Add a subcommand and return the input function unchanged. Suitable to use as decorator."""
        if func is not None:
            self.new_subcommand(func)
            return func

        def wrap(func):
            self.new_subcommand(func, **kwargs)
            return func

        return wrap

    def add_subcommand(self, func: Callable[..., Any], *args, **kwargs) -> Self:
        """Add a subcommand and return the caller object. Suitable to add multiple subcommands in a row."""
        self.new_subcommand(func, *args, **kwargs)
        return self

    def end_subcommand(self, func: Callable[..., Any], *args, **kwargs) -> Command:
        """Add a subcommand and return the parent `Command` instance of the caller object.
        If `parent` attribute is `None`, raise `ValueError`."""
        if self.parent is None:
            raise ValueError(
                "\n\nMethod `end_subcommand()` can not be called by `Command` instances without parent.\n\n"
            )
        self.new_subcommand(func, *args, **kwargs)
        return self.parent

    def new_subcommand(
        self,
        func: Callable[..., Any],
        name: str | None = None,
        help: str | None = None,
        aliases: Sequence[str] | None = None,
        *args,
        **kwargs,
    ) -> Command:
        """Add a subcommand and return the new created subcommand (a new `Command` instance)"""
        # TODO: add `deprecated` included in v3.13
        count = 0
        parent_parser = self.parent
        while parent_parser:
            count += 1
            parent_parser = parent_parser.parent
        cmd: Command = Command(func, *args, **kwargs)
        cmd.name = name or func.__name__
        if not hasattr(self, "subparsers_dest"):
            self.subparsers_dest: str = ""
        self.subparsers_dest = ",".join(
            [self.subparsers_dest[1 : -1 * (count + 1)], cmd.name] if self.subparsers_dest else [cmd.name]
        )
        self.subparsers_dest = "{" + self.subparsers_dest + "}" + " " * count
        cmd.aliases = aliases or []
        cmd.help = help
        cmd.parent = self
        cmd.__sanitize_argument_data_names()
        self.sub_commands.update({cmd.name: cmd})
        return cmd

    def run(self, args: Sequence[str] | None = None) -> Any:
        # TODO: `Context` object
        # TODO: treat variatic argument as parse_know
        # TODO: treat "positional only"?
        if args == None:
            args = sys.argv[1:]
        if self.parser is None:
            self._add_parsers()
        assert self.parser is not None
        namespace: Namespace = self.parser.parse_args(args)
        # TODO Enum decoverter
        for arg in self.argument_data:
            if isinstance(arg.typeannotation, type) and issubclass(arg.typeannotation, Enum):
                setattr(namespace, arg.name, arg.typeannotation[getattr(namespace, arg.name)])
            if get_origin(arg.typeannotation) is Literal:
                types = get_args(arg.typeannotation)
                for t in types:
                    choice_type = type(t)
                    if issubclass(choice_type, Enum):
                        try:
                            setattr(namespace, arg.name, choice_type[getattr(namespace, arg.name)])
                        except:
                            continue
        subcommand_name = (
            getattr(namespace, self.subparsers_dest) if hasattr(self, "subparsers_dest") else None
        )
        result = None
        if self.func:
            result = self.func(**{k: _getattr_with_spaces(namespace, k) for k in self.parameters})
        if subcommand_name is not None:
            args = args[args.index(subcommand_name) + 1 :]
            return self.sub_commands[subcommand_name].run(args)
        return result

    ##########################################################################################################
    # %:          PRIVATE METHODS
    ##########################################################################################################

    def __generate_argument_data_list(self) -> list[_ArgumentData]:
        argument_data: list[_ArgumentData] = []
        for par in self.parameters:
            data: _ArgumentData = _get_argument_data_from_parameter(self.parameters[par])
            data.help = self.docstring_data.helps.get(data.name, None) if self.docstring_data else None
            argument_data.append(data)
        return argument_data

    def __sanitize_argument_data_names(self) -> None:
        if self.parent:
            names: list[str] = [arg.name for arg in self.parent.argument_data]
            strip_names: list[str] = [n.strip() for n in names]
            for arg in self.argument_data:
                if arg.name.strip() in strip_names:
                    arg.name = names[strip_names.index(arg.name.strip())] + " "

    def _get_data_from_docstring(self) -> _DocstringData | None:
        if self.docstring_template:
            return self._collect_docstring_data_using_template(self.docstring_template)
        for template in DOCSTRING_TEMPLATES:
            data: _DocstringData | None = self._collect_docstring_data_using_template(template)
            if data:
                return data
        return None

    def _collect_docstring_data_using_template(self, template: str | None = None) -> _DocstringData | None:
        separator: str = "################################" * 30
        template = template or self.docstring_template
        parameter_number = len(self.parameters)
        docstring = _normalize_docstring(self.func.__doc__)
        # escape for regex match, but not "{" and "}"
        template = re.escape(_normalize_docstring(template) + "\n").replace(r"\{", "{").replace(r"\}", "}")
        place_holders: dict[str, list[int]] = {
            "description": [],
            "epilog": [],
            "parameter_name": [],
            "parameter_type": [],
            "parameter_description": [],
        }
        detected_place_holders: list[str] = re.findall(r"{{.*?}}", template)
        order_counter = 0
        for word in detected_place_holders:
            word = word.removeprefix("{{").removesuffix("}}")
            if word in place_holders:
                place_holders[word].append(order_counter)
                order_counter += 1
        parameter_section_init_index: int = 0
        for i, line in enumerate(template.splitlines()):
            if any([f"{{{{{key}}}}}" in line for key in place_holders if key.startswith("parameter")]):
                parameter_section_init_index = i
                break
        parameter_section_length = sum(
            [template.count(f"{{{{{key}}}}}") for key in place_holders if key.startswith("parameter")]
        )
        if parameter_number > 0 and not parameter_section_length:
            return None
        if parameter_section_length:
            parameter_section = "\n".join(template.splitlines()[parameter_section_init_index:])
            for _ in range(parameter_number - 1):
                template += f"{parameter_section}\n"
        else:
            docstring = docstring.rstrip() + f"\n{separator}"
            template = template.rstrip() + f"\n{re.escape(separator)}\n"
        for place_holder in place_holders:
            template = template.replace(f"{{{{{place_holder}}}}}", "(?! )(.*?)")
        docstring += "\n"
        template += "(?!\\s)(.*?)"
        match = re.match(template, docstring, re.DOTALL)
        if match:
            matches: tuple[str, ...] = match.groups()
            description = matches[place_holders["description"][0]] if place_holders["description"] else None
            epilog = matches[place_holders["epilog"][0]] if place_holders["epilog"] else None
            docstring_data = _DocstringData(description=description, epilog=epilog)
            for i in range(parameter_number):
                docstring_data.helps[
                    matches[place_holders["parameter_name"][0] + parameter_section_length * i]
                ] = _normalize_docstring(
                    matches[place_holders["parameter_description"][0] + parameter_section_length * i].strip()
                )
            return docstring_data
        return None

    def __make_argflagged(self, name: str) -> str:
        return f"{self.longstartflags}{name.replace("_","-")}"

    def __doesnothavelongstartflag(self, flags: Sequence[str]) -> bool:
        return not any([flag.startswith(f"{self.longstartflags}") for flag in flags])

    def _generate_args_for_add_argument(
        self, argdata: _ArgumentData
    ) -> tuple[tuple[str, ...], _CompleteKeywordArguments]:
        """Helper function to get data from the proxy object and creates (args, kwargs) to `add_argument()`
        Ref: https://docs.python.org/3/library/argparse.html#the-add-argument-method
        """
        # TODO: check variadic args and kwargs
        kwargs: _CompleteKeywordArguments = {
            "dest": argdata.name,
            "help": argdata.kwargs.get("help", argdata.help),
            "default": argdata.kwargs.get("default", argdata.default),
        }
        default_bool = kwargs["default"] if kwargs["default"] is not EMPTY else self.default_bool
        action, nargs, argtype, choices = "store", None, str, None
        if argdata.typeannotation is not None:
            action, nargs, argtype, choices = _get_data_from_typeannotation(
                argdata.typeannotation, default_bool
            )
        kwargs["action"] = argdata.kwargs.get("action") or action
        if kwargs["action"] in ["store", "append"]:
            kwargs["type"] = argdata.kwargs.get("type") or argtype
            kwargs["nargs"] = argdata.kwargs.get("nargs") or nargs
            kwargs["choices"] = argdata.kwargs.get("choices") or choices
        argdata.make_flag = (
            all(
                [
                    argdata.make_flag is None,
                    self.__doesnothavelongstartflag(argdata.flags),
                    kwargs["default"] is not EMPTY,
                    kwargs.get("nargs") not in ["*", "?"],
                ]
            )
            or kwargs["action"] in ["store_true", "store_false"]
            or argdata.make_flag
        )
        argflagged: str | None = None
        if argdata.make_flag or all(
            [
                argdata.make_flag is None,
                argdata.flags,
                self.__doesnothavelongstartflag(argdata.flags),
            ]
        ):
            argflagged = self.__make_argflagged(argdata.name)
        if argflagged:
            argdata.flags.append(argflagged)
        if kwargs["default"] is EMPTY:
            kwargs["default"] = None
            if argdata.flags:
                kwargs["required"] = argdata.kwargs.get("required") or True
                if kwargs["action"] in ["store_true", "store_false"]:
                    kwargs["action"] = BooleanOptionalAction

        # given in `argdata.kwargs` has preference over inferred
        for key in ["metavar", "const", "version"]:
            try:
                kwargs[key] = argdata.kwargs.pop(key)  # type: ignore
            except KeyError:
                pass
        return tuple(argdata.flags), kwargs

    def _add_parsers(self) -> None:
        if self.parent is None:
            self.parser = ArgumentParser(
                prog=self.prog or self.func.__name__ if self.func else None,
                usage=self.usage,
                description=self.description,
                epilog=self.epilog,
                parents=self.parents,
                formatter_class=self.formatter_class,
                prefix_chars=self.prefix_chars,
                fromfile_prefix_chars=self.fromfile_prefix_chars,
                argument_default=self.argument_default,
                conflict_handler=self.conflict_handler,
                add_help=self.add_help,
                allow_abbrev=self.allow_abbrev,
                exit_on_error=self.exit_on_error,
            )
        else:
            assert self.parent.sub_commands_group and self.name
            self.parser = self.parent.sub_commands_group.add_parser(
                name=self.name,
                help=self.help,
                aliases=self.aliases,
                prog=self.prog,
                usage=self.usage,
                description=self.description,
                epilog=self.epilog,
                parents=self.parents,
                formatter_class=self.formatter_class,
                prefix_chars=self.prefix_chars,
                fromfile_prefix_chars=self.fromfile_prefix_chars,
                argument_default=self.argument_default,
                conflict_handler=self.conflict_handler,
                add_help=self.add_help,
                allow_abbrev=self.allow_abbrev,
                exit_on_error=self.exit_on_error,
            )
        self.arguments: list[Action] = []
        for argument_data in self.argument_data:
            flags, kwargs = self._generate_args_for_add_argument(argument_data)
            self.arguments.append(self.parser.add_argument(*flags, **kwargs))  # type:ignore

        assert self.parser is not None
        if self.sub_commands and not self.sub_commands_group:
            # ref: https://docs.python.org/3/library/argparse.html#argparse.ArgumentParser.add_subparsers
            self.sub_commands_group = self.parser.add_subparsers(
                title=self.subcommands_title,
                description=self.subcommands_description,
                prog=self.subcommands_prog,  # I corrected this, see https://github.com/python/typeshed/issues/13162
                required=self.subcommands_required,
                help=self.subcommands_help,
                metavar=self.subcommands_metavar,
                dest=self.subparsers_dest,
            )

        for cmd in self.sub_commands:
            self.sub_commands[cmd]._add_parsers()

    @property
    def is_main_command(self) -> bool:
        return self.parent is None

    # endregion


##############################################################################################################
# %%          PRIVATE CLASSES
##############################################################################################################


class _ArgumentMetaDataDictionary(TypedDict, total=False):
    """Dictionary with some parameters passed to the original `add_argument()` method.
    These are expected to be in the argument metadata annotation.
    Namely: `action`, `nargs`, `const`, `choices`, `required`, `help`, `metavar`, `version`.
    The parameter `version` is not documented, but is on some stub.
    The parameters `name_or_flags`, `default`, `type`, `dest` are not passed in this dictionary.
    Ref: https://docs.python.org/3/library/argparse.html#the-add-argument-method
    """

    action: (
        str
        | Literal[
            "store",
            "store_const",
            "store_true",
            "store_false",
            "append",
            "append_const",
            "count",
            "help",
            "version",
            "extend",
        ]
        | type[Action]
    )
    nargs: int | str | None
    const: Any
    choices: Iterable | None
    required: bool | None
    help: str | None
    metavar: str | tuple[str, ...] | None
    version: str | None


class KeywordArguments(_ArgumentMetaDataDictionary, total=False):
    """Dictionary inheriting parameters passed to the original `add_argument()` method,
    including `default` and `type`. These are suppose to be passed to the `add_argument()` method, after
    including `dest` and `name_or_flags`.
    Ref: https://docs.python.org/3/library/argparse.html#the-add-argument-method
    """

    default: Any
    type: type | Callable[[str], Any] | None


class _CompleteKeywordArguments(KeywordArguments, total=False):
    """Dictionary with all parameters passed to the original `add_argument()` method,
    including `dest` . These are suppose to be passed to the `add_argument()` method after
    including `name_or_flags`, which is positional (not a keyword argument).
    Ref: https://docs.python.org/3/library/argparse.html#the-add-argument-method
    """

    dest: str


@dataclass
class _DocstringData:
    """A dataclass with data recovered from docstring"""

    description: str | None
    epilog: str | None
    helps: dict[str, str] = field(default_factory=dict)


@dataclass
class _ArgumentData:
    """A proxy dataclass to store info that came from `inspect.Parameter` objects
    Ref: https://docs.python.org/3/library/inspect.html#inspect.Parameter
    """

    name: str
    typeannotation: Callable[[str], Any] | str | FileType | None = None
    kind: Kind = Kind.POSITIONAL_OR_KEYWORD
    default: Any = Parameter.empty
    flags: list[str] = field(default_factory=list)
    kwargs: KeywordArguments = field(default_factory=KeywordArguments)
    make_flag: bool | None = None
    argument_group: ArgumentGroup | None = None
    mutually_exclusive_group: MutuallyExclusiveGroup | None = None
    parser: Any = None
    help: str | None = None


##############################################################################################################
# %%          PUBLIC CLASSES
##############################################################################################################


@dataclass
class ArgumentGroup:
    title: str | None = None
    description: str | None = None
    _: KW_ONLY
    argument_default: Any = None
    conflict_handler: str = "error"


@dataclass
class MutuallyExclusiveGroup:
    required: bool = False
    _: KW_ONLY
    argument_group: ArgumentGroup | None = None


@dataclass
class ArgumentMetaData:
    flags: list[str] = field(default_factory=list)
    make_flag: bool | None = None
    argument_group: ArgumentGroup | None = None
    mutually_exclusive_group: MutuallyExclusiveGroup | None = None
    dictionary: KeywordArguments = field(default_factory=KeywordArguments)


##############################################################################################################
# %%          PRIVATE FUNCTIONS
##############################################################################################################


def _getattr_with_spaces(namespace: Namespace, name: str) -> Any:
    """Like `getattr` but try to get attributes with spaces appended to the names"""
    # Try exact match first
    if hasattr(namespace, name):
        return getattr(namespace, name)

    # Otherwise, keep appending spaces until found
    padded = name
    while True:
        padded += " "
        if hasattr(namespace, padded):
            return getattr(namespace, padded)


def _normalize_docstring(docstring: str | None) -> str:
    """https://peps.python.org/pep-0257/#handling-docstring-indentation

    This functions maybe do the same as `inspect.cleandoc`.
    However, this one accepts `None`.
    """
    if not docstring:
        return ""
    lines: list[str] = docstring.expandtabs(tabsize=4).splitlines()
    indentation: int = min([len(line) - len(line.lstrip()) for line in lines[1:] if line.lstrip()], default=0)
    lines: list[str] = [lines[0].strip()] + [
        line.removeprefix(" " * indentation).rstrip() for line in lines[1:]
    ]
    while lines and not lines[-1]:
        lines.pop()
    while lines and not lines[0]:
        lines.pop(0)
    return "\n".join(lines)


def _get_argument_data_from_parameter(parameter: Parameter) -> _ArgumentData:
    """Helper function to get data from a `inspect.Parameter` object and generetes a proxy object
    Ref: https://docs.python.org/3/library/inspect.html#inspect.Parameter
    """
    data: _ArgumentData = _ArgumentData(name=parameter.name, kind=parameter.kind)
    data.default = parameter.default
    if parameter.annotation is not EMPTY:
        annotation = parameter.annotation
        if type(annotation) == str:
            annotation = eval(annotation)
        data.typeannotation = annotation
        if hasattr(annotation, "__metadata__"):
            data.typeannotation = annotation.__origin__
            metadatas = annotation.__metadata__
            for metadata in metadatas:
                if isinstance(metadata, ArgumentMetaData):
                    data.flags = metadata.flags
                    data.make_flag = metadata.make_flag
                    data.argument_group = metadata.argument_group
                    data.mutually_exclusive_group = metadata.mutually_exclusive_group
                    data.kwargs = metadata.dictionary
                    break
    if parameter.annotation is EMPTY and parameter.default is not EMPTY:
        data.typeannotation = type(parameter.default)
    return data


def _get_data_from_typeannotation(
    annotation: Any,
    default_bool: bool = False,
) -> tuple[str, str | int | None, type | Callable[[str], Any] | None, Sequence[Any] | None]:
    """Return `action`, `nargs`, `argtype`, `choices`"""
    action = "store"
    nargs = None
    argtype = annotation if callable(annotation) else str
    choices = None
    origin = get_origin(annotation)
    if origin:
        types = get_args(annotation)
        if origin in [Union, UnionType]:
            types = [t for t in get_args(annotation) if t is not type(None)]
            argtype = __create_union_converter(types)
        if origin is tuple:
            nargs = len(types) if Ellipsis not in types else "*"
            argtype = types[0]
        if origin in [list, Sequence]:
            nargs = "*"
            argtype = types[0]
        if origin is Literal:
            choices = [t.name if isinstance(t, Enum) else t for t in types]
            argtype = None  # create_literal_converter(types)
    if annotation == bool:
        action = "store_false" if default_bool else "store_true"
        argtype = None
    if isinstance(argtype, type) and issubclass(argtype, Enum):
        choices = list(getattr(argtype, "__members__").keys())
        argtype = None

    return action, nargs, argtype, choices


def __create_union_converter(types):
    if len(types) == 1 and issubclass(types[0], Enum):
        return types[0]

    def converter(value: str) -> Any:
        for t in types:
            try:
                if issubclass(t, Enum):
                    return t[value]
                # Attempt conversion
                converted_value = t(value)
                # Check string representation matches
                if str(converted_value) == value:
                    return converted_value
            except (ValueError, TypeError):
                continue  # Ignore and try the next type
        raise ValueError("ERRO")

    return converter


##############################################################################################################
# %%          UNUSED FUNCTIONS
##############################################################################################################


def __create_literal_converter(types):
    def converter(s):
        for value in types:
            if isinstance(value, Enum) and s == getattr(value, "name"):
                return getattr(value, "name")
            if str(value) == s:
                return value
        raise ValueError("ERRO")

    return converter


def __count_leading_spaces(string: str):
    return len(string) - len(string.lstrip())


def __arg(
    *flags: str,
    make_flag: bool | None = None,
    group: ArgumentGroup | None = None,
    mutually_exclusive_group: MutuallyExclusiveGroup | None = None,
    subparser: Field[Any] | None = None,
    **kwargs: Unpack[KeywordArguments],
) -> Any:
    """"""
    return field(
        default=kwargs.pop("default", None),
        metadata={
            "obj": ArgumentMetaData(
                flags=list(flags),
                make_flag=make_flag,
                argument_group=group,
                mutually_exclusive_group=mutually_exclusive_group,
                dictionary=kwargs,
            ),
            "subparser": subparser,
        },
    )


def __get_metadata_from_field(field: Field[Any]) -> _ArgumentData:
    if type(field.type) == str:
        field.type = eval(field.type)
    data: _ArgumentData = _ArgumentData(name=field.name, typeannotation=field.type)
    if field.default is not field.default_factory:
        data.default = field.default
    if field.metadata:
        data.parser = field.metadata.get("subparser", None)
        metadata: ArgumentMetaData = field.metadata.get("obj", None)
        data.flags = metadata.flags
        data.make_flag = metadata.make_flag
        data.argument_group = metadata.argument_group
        data.mutually_exclusive_group = metadata.mutually_exclusive_group
        data.kwargs = metadata.dictionary
    return data


##############################################################################################################
# %%          PUBLIC FUNCTIONS
##############################################################################################################


def data(
    *flags: str,
    make_flag: bool | None = None,
    group: ArgumentGroup | None = None,
    mutually_exclusive_group: MutuallyExclusiveGroup | None = None,
    **kwargs: Unpack[KeywordArguments],
) -> ArgumentMetaData:
    return ArgumentMetaData(
        flags=list(flags),
        make_flag=make_flag,
        argument_group=group,
        mutually_exclusive_group=mutually_exclusive_group,
        dictionary=kwargs,
    )


def run(func: Callable[..., Any], args: Sequence[str] | None = None, **kwargs):
    Command(func, **kwargs).run(args)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "clig"
copyright = "2024, Diogo Rossi"
author = "Diogo Rossi"
release = "0.1.0"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

import os
from pathlib import Path


import sys
from pathlib import Path

path = Path(__file__).parent
sys.path.insert(0, str((path).resolve()))
os.chdir(path)

import convert_notebook

extensions = ["myst_parser", "sphinx_copybutton"]

templates_path = ["_templates"]
exclude_patterns = []


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "furo"
html_static_path = ["_static"]



# cSpell: disable
import sys
from pathlib import Path
from argparse import BooleanOptionalAction

this_dir = Path(__file__).parent

sys.path.insert(0, str((this_dir).resolve()))
sys.path.insert(0, str((this_dir / "../src").resolve()))
sys.path.insert(0, str((this_dir / "../src").resolve()))
from clig.clig import Command, _normalize_docstring, _CompleteKeywordArguments
import functions as fun


def test_inferarg__pn_knc_noDoc():
    cmd = Command(fun.pn_knc_noDoc)
    arg_1, arg_2 = cmd.argument_data
    assert cmd._generate_args_for_add_argument(arg_1) == (
        (),
        _CompleteKeywordArguments(
            action="store",
            dest="first",
            type=str,
            default=None,
            nargs=None,
            choices=None,
            help=None,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_2) == (
        ("--second",),
        _CompleteKeywordArguments(
            action="store",
            dest="second",
            type=str,
            default="test",
            nargs=None,
            choices=None,
            help=None,
        ),
    )


def test_inferarg__pn_pt_kti_noDoc():
    cmd = Command(fun.pn_pt_kti_noDoc)
    arg_a, arg_b, arg_c = cmd.argument_data
    assert cmd._generate_args_for_add_argument(arg_a) == (
        (),
        _CompleteKeywordArguments(
            action="store",
            dest="a",
            type=str,
            default=None,
            nargs=None,
            choices=None,
            help=None,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_b) == (
        (),
        _CompleteKeywordArguments(
            action="store",
            dest="b",
            type=float,
            default=None,
            nargs=None,
            choices=None,
            help=None,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_c) == (
        ("--c",),
        _CompleteKeywordArguments(
            action="store",
            dest="c",
            type=int,
            default=123,
            nargs=None,
            choices=None,
            help=None,
        ),
    )


def test_inferarg__ptc_kti_ktb_cligDocMutiline():
    cmd = Command(fun.ptc_kti_ktb_cligDocMutiline)
    arg_a, arg_b, arg_c = cmd.argument_data
    assert cmd._generate_args_for_add_argument(arg_a) == (
        (),
        _CompleteKeywordArguments(
            action="store",
            dest="a",
            type=str,
            default=None,
            nargs=None,
            choices=None,
            help=arg_a.help,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_b) == (
        ("--b",),
        _CompleteKeywordArguments(
            action="store",
            dest="b",
            type=int,
            default=123,
            nargs=None,
            choices=None,
            help=arg_b.help,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_c) == (
        ("--c",),
        _CompleteKeywordArguments(
            action="store_false",
            dest="c",
            default=True,
            help=arg_c.help,
        ),
    )


def test_inferarg__ptcm_ptcm_ktb():
    cmd = Command(fun.ptcm_ptcm_ktb)
    arg_a, arg_b, arg_c = cmd.argument_data
    assert cmd._generate_args_for_add_argument(arg_a) == (
        ("-f", "--first"),
        _CompleteKeywordArguments(
            dest="a",
            default=None,
            nargs=None,
            required=True,
            choices=None,
            action="store",
            type=str,
            help="The first argument",
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_b) == (
        (),
        _CompleteKeywordArguments(
            dest="b",
            action="store_const",
            default=None,
            const=123,
            help=None,
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_c) == (
        ("--c",),
        _CompleteKeywordArguments(
            dest="c",
            help=None,
            default=True,
            action="store_false",
        ),
    )


def test_inferarg__pti_ptc_ptf_ktb_ktlo_numpyEpilogMultiline():
    cmd = Command(fun.pti_ptc_ptf_ktb_ktlo_numpyEpilogMultiline)
    args_a, args_b, args_c, args_d, args_e = cmd.argument_data
    assert cmd._generate_args_for_add_argument(args_a) == (
        (),
        _CompleteKeywordArguments(
            dest="a",
            default=None,
            action="store",
            type=int,
            nargs=None,
            choices=None,
            help=_normalize_docstring(
                """Fuga nemo provident vero odio qui sint et aut veritatis. Facere necessitatibus ut. Voluptatem
                natus natus veritatis earum. Reprehenderit voluptate dolorem dolores consequuntur magnam impedit
                eius. Est ut nisi aut accusamus."""
            ),
        ),
    )
    assert cmd._generate_args_for_add_argument(args_b) == (
        (),
        _CompleteKeywordArguments(
            dest="b",
            default=None,
            action="store",
            type=str,
            nargs=None,
            choices=None,
            help="Culpa asperiores incidunt molestias aliquam soluta voluptas excepturi nulla.",
        ),
    )


def test_inferarg__ptc_ptb_cligEpilog():
    cmd = Command(fun.ptc_ptb_cligEpilog)
    arg_name, arg_flag = cmd.argument_data
    assert cmd._generate_args_for_add_argument(arg_name) == (
        (),
        _CompleteKeywordArguments(
            dest="name",
            default=None,
            action="store",
            type=str,
            nargs=None,
            choices=None,
            help="Sequi deserunt est quia qui.",
        ),
    )
    assert cmd._generate_args_for_add_argument(arg_flag) == (
        ("--flag",),
        _CompleteKeywordArguments(
            dest="flag",
            default=None,
            action=BooleanOptionalAction,
            required=True,
            help="Labore eius et voluptatem quos et consequatur dolores.",
        ),
    )



# cSpell: disable
import inspect
import sys
from pathlib import Path

this_dir = Path(__file__).parent

sys.path.insert(0, str((this_dir).resolve()))
sys.path.insert(0, str((this_dir / "../src").resolve()))

from clig.clig import _get_argument_data_from_parameter, _ArgumentData, Kind
import functions as fun


def test_parameter_without_annotation():
    parameter = inspect.signature(fun.pn_noDoc).parameters["a"]
    argmetadata = _get_argument_data_from_parameter(parameter)
    assert argmetadata == _ArgumentData(name="a", kind=Kind.POSITIONAL_OR_KEYWORD)


def test_get_argdata_from_parameter_posWithType_posBoolWithType_cligDoc():
    parameters = inspect.signature(fun.ptc_ptb_cligEpilog).parameters
    par_name, par_flag = parameters["name"], parameters["flag"]
    arg_data_par_name = _get_argument_data_from_parameter(par_name)
    arg_data_par_flag = _get_argument_data_from_parameter(par_flag)
    assert arg_data_par_name == _ArgumentData(name="name", typeannotation=str)
    assert arg_data_par_flag == _ArgumentData(name="flag", typeannotation=bool)



# cSpell: disable
import sys
from pathlib import Path

this_dir = Path(__file__).parent

sys.path.insert(0, str((this_dir).resolve()))
sys.path.insert(0, str((this_dir / "../src").resolve()))
from clig import clig  # protected functions
import functions as fun


def adjust_epilog_for_test(text: str) -> str:
    """First and last lines must be always empty"""
    return "\n".join([line.strip() for line in text.splitlines()[1:-1]])


def test_docstringdata__descr():
    cmd = clig.Command(fun.descr, docstring_template=clig.DESCRIPTION_DOCSTRING)
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "A foo that bars"


def test_docstringdata__descrEpilog_fromFun():
    cmd = clig.Command(fun.descrEpilog, docstring_template=clig.DESCRIPTION_DOCSTRING)
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == clig._normalize_docstring(fun.descrEpilog.__doc__)


def test_docstringdata__descrEpilog():
    epilog = adjust_epilog_for_test(
        """
        Corporis ullam nam ut dolores sed. Nemo ea deserunt facere numquam velit aut. Architecto provident
        consequatur ratione est quas qui dolor ratione. Laudantium fugit at.

        Ullam et temporibus eum. Sit voluptatem tempora totam dolores. Pariatur accusamus voluptate totam.
        Fugit rerum nemo reiciendis veritatis modi sit distinctio ratione id.

        Voluptates tenetur quos qui exercitationem laudantium aliquid. Neque qui eum qui. Qui tenetur facilis
        non voluptatem ut corporis harum fugiat.
        """
    )
    cmd = clig.Command(fun.descrEpilog, docstring_template=clig.DESCRIPTION_EPILOG_DOCSTRING)
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Aliquam alias quia earum."
    assert data.epilog == epilog


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_numpyEpilog():
    epilog = adjust_epilog_for_test(
        """
        Blanditiis velit consequatur omnis odit magnam quo dignissimos. Qui ex et illo. Et
        necessitatibus ea placeat consectetur itaque dolore fugiat quo autem. Ut accusamus incidunt
        repellat minima est soluta ut est. Id aut enim ad. Quia qui sint ex eos eveniet eveniet
        earum unde.

        Non rerum aut consectetur ut ducimus ut similique ut illum. Aut qui et distinctio. Nihil id
        sit incidunt minus omnis. Quo unde inventore fuga. Quasi ea ea dolores quam.

        Esse temporibus voluptas nulla. Odio voluptas nisi quae cupiditate consequatur cumque ut ex
        dolorem. Est doloremque quis nostrum voluptates doloremque quia. Ex sunt dolores consectetur
        veritatis maxime suscipit. Fugit dolorem facilis quasi.
        """
    )

    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_numpyEpilog,
        docstring_template=clig.NUMPY_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Distinctio et ratione sequi hic."
    assert data.epilog == epilog
    assert all([p in d for p, d in zip(["a", "b", "c", "d", "e"], data.helps.keys())])
    assert data.helps["a"] == "Neque ut qui non nulla odit esse accusantium aut suscipit."
    assert data.helps["b"] == "Nihil dolores autem autem nulla sit nihil molestiae vero est."
    assert data.helps["c"] == "Animi magnam ut sapiente maiores."
    assert data.helps["d"] == "Consequatur provident neque optio consequatur., by default True"
    assert data.helps["e"] == "Corrupti molestiae in aspernatur., by default None"


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_sphinxEpilog():
    epilog = adjust_epilog_for_test(
        """
        Est nam quia voluptatem vero architecto laborum. Accusantium delectus et aut repudiandae
        voluptatibus qui iure ut debitis. Voluptatibus ut enim consequatur iusto eaque dolor.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_sphinxEpilog,
        docstring_template=clig.SPHINX_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Qui accusantium harum debitis et."
    assert data.epilog == epilog
    assert data.helps["a"] == "Atque pariatur excepturi sed dolorem sint impedit molestiae."
    assert data.helps["b"] == "magni optio voluptatibus"
    assert data.helps["c"] == "Sit eligendi consequatur recusandae doloribus enim amet."
    assert data.helps["d"] == "Soluta dolorum amet et., defaults to True"
    assert data.helps["e"] == "Fugiat provident amet iste natus ab voluptas., defaults to None"


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_googleEpilog():
    epilog = adjust_epilog_for_test(
        """
        Maiores occaecati quam asperiores non sunt est dolor laborum est. Eius corporis nobis
        accusamus rerum et et et. Ducimus tempore voluptas qui aut consectetur saepe quos cum
        delectus. Tempora adipisci odit qui. Optio eum magni non. Placeat repudiandae quasi nostrum
        mollitia sunt neque fuga id possimus.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_googleEpilog,
        docstring_template=clig.GOOGLE_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Voluptatum dolorem quis dolorum voluptas atque non temporibus."
    assert data.epilog == epilog
    assert data.helps["a"] == "Qui eum eius nihil voluptas quia aut numquam."
    assert data.helps["b"] == "Quasi voluptates dicta cumque similique qui dolorem architecto."
    assert data.helps["c"] == "Sint harum et omnis nobis numquam quos omnis."
    assert data.helps["d"] == "Ex voluptas animi.. Defaults to True."
    assert data.helps["e"] == "In vero ut nisi officia ut.. Defaults to None."


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_cligEpilog():
    clig_example_epilog = adjust_epilog_for_test(
        """
        Neque dolores expedita repellat in perspiciatis dolorem aliquid et. Commodi fugit minima
        laudantium beatae et ut. Id possimus soluta magnam quisquam laboriosam impedit.

        Ad quaerat ut culpa aut iure id quia. Ut aut alias adipisci quia. Veritatis ratione
        dignissimos laborum. Molestiae molestias id earum.

        Nesciunt quas corrupti tenetur officiis occaecati asperiores eaque. Qui voluptas ut ea dolor
        et harum beatae quos. Est tenetur ut ipsum. Eveniet rem beatae error eum voluptatem tempora
        velit in. Ea doloribus similique.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_cligEpilog,
        docstring_template=clig.CLIG_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Fugit voluptatibus enim odit velit facilis."
    assert data.epilog == clig_example_epilog
    assert data.helps["a"] == "Quidem natus sunt molestiae et reprehenderit voluptas optio."
    assert data.helps["b"] == "Unde rerum aut a et assumenda fugit dolorem eligendi corrupti."
    assert data.helps["c"] == "Dolorum officiis totam aspernatur fuga voluptas similique."
    assert data.helps["d"] == "Ducimus sunt eum in vel voluptatibus aut facere perspiciatis."
    assert data.helps["e"] == "Sit et consequatur a asperiores sequi sint dolores id ipsam."


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_numpyEpilogMultiline():
    epilog = adjust_epilog_for_test(
        """
        Qui deserunt sequi aut illo. Minima modi illo sit occaecati. Ducimus illum et. Deleniti repellendus
        cum quasi ut et natus dolorem. Aliquam aut aperiam qui.

        Sint maiores dolorum. Nobis quo distinctio consequuntur. Recusandae fuga qui perspiciatis quisquam.
        Nostrum asperiores neque nisi. Enim voluptatem eum illo.

        Est labore illum voluptatibus at ut. Deleniti ut aut ut aperiam consequatur aut. Reprehenderit
        voluptatem est voluptates temporibus et voluptate accusamus dolores. Placeat nihil dignissimos sed
        sequi sequi.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_numpyEpilogMultiline,
        docstring_template=clig.NUMPY_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Voluptatibus eos ipsa ex debitis voluptatem dignissimos."
    assert data.epilog == epilog
    assert data.helps["a"] == clig._normalize_docstring(
        """Fuga nemo provident vero odio qui sint et aut veritatis. Facere necessitatibus ut. Voluptatem
    natus natus veritatis earum. Reprehenderit voluptate dolorem dolores consequuntur magnam impedit
    eius. Est ut nisi aut accusamus."""
    )
    assert data.helps["b"] == clig._normalize_docstring(
        """Culpa asperiores incidunt molestias aliquam soluta voluptas excepturi nulla."""
    )
    assert data.helps["c"] == clig._normalize_docstring(
        """Non vitae qui non magni harum eum maiores qui. Dicta sunt voluptatem voluptate. At quibusdam
    aliquam autem et perspiciatis et assumenda. Perferendis qui velit quam numquam iste."""
    )
    assert data.helps["d"] == clig._normalize_docstring(
        """Aut ipsam aut velit impedit. Quidem expedita aliquid sed officia in ex et. Nihil rem adipisci ut
    perferendis iure., by default True"""
    )
    assert data.helps["e"] == clig._normalize_docstring(
        """Ratione consequatur molestiae quia deserunt quo. Non cupiditate sunt commodi vero labore
    doloremque voluptatem officiis est. Iusto voluptate reiciendis iusto in. Occaecati quia soluta
    minus perspiciatis alias illum iste aperiam et. Autem accusamus unde omnis est cum ducimus. Iure
    adipisci id omnis quis placeat impedit rerum ab aspernatur.

    Praesentium id rerum quod provident odit dolores adipisci veniam natus. Porro repellat aliquid
    quibusdam recusandae hic voluptas accusantium voluptatem voluptatem. Laboriosam similique nobis
    aut iusto et ab minima cum.

    Voluptas molestiae mollitia autem distinctio magnam dolorem molestiae aliquid. Neque provident
    impedit et. Quod quibusdam nulla cupiditate. Praesentium neque vel ea velit consequatur quis
    voluptate iste. Quae veniam sequi et nihil qui vel voluptatem maxime. Laborum corrupti dolores
    voluptate placeat fugit non nobis., by default None"""
    )


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_sphinxEpilogMultiline():
    epilog = adjust_epilog_for_test(
        """
        Quia aspernatur doloribus id doloribus sunt ratione et voluptatum. Eligendi numquam sed. Voluptas
        consequuntur quibusdam debitis quia unde doloribus ducimus sunt. Et provident assumenda hic eum sint
        quia ipsum sit sed.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_sphinxEpilogMultiline,
        docstring_template=clig.SPHINX_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Est sit minus quasi soluta unde vero deleniti eligendi."
    assert data.epilog == epilog
    assert data.helps["a"] == """Velit ratione harum in quia laborum ut est quis."""
    assert data.helps["b"] == clig._normalize_docstring(
        """Adipisci voluptates aut fugiat qui nam non. Eveniet molestiae voluptas explicabo fuga.
        Beatae ex sed nostrum incidunt."""
    )
    assert data.helps["c"] == clig._normalize_docstring(
        """Non non voluptatum ipsum sit maiores et eum adipisci. Autem sit possimus et similique atque.
        Nihil tempore et excepturi.

        Nisi magnam et. Illum minus ea enim eligendi doloremque consequatur odit est officiis. Dolorem
        dolores repellat esse vero quae. Laboriosam ab qui quo eveniet quia ex et. Aut facilis molestias
        qui. Dolorum sit magni repellat iusto aut vel.

        Laborum dolores illum modi. Id et qui nisi harum aperiam doloribus. Quod quibusdam dolorum iusto."""
    )
    assert data.helps["d"] == clig._normalize_docstring(
        """Et magni harum adipisci accusantium aut et ipsum impedit. Sit modi voluptatem. Esse quis aut
        ex. Dicta quam rem repellendus accusantium aut molestias praesentium fugiat corporis. Assumenda
        eum natus voluptatem alias dolorem vitae dolor repudiandae inventore. Et deleniti repellendus
        quo., defaults to True"""
    )
    assert data.helps["e"] == clig._normalize_docstring(
        """Corporis est rerum. Aspernatur dolor porro a culpa omnis. Repudiandae totam necessitatibus
        quibusdam ipsum numquam eveniet dolor quasi. Dolores dolorem voluptate aut. Deleniti officia qui
        molestiae. Quo deserunt nulla aut sit sunt quam nostrum odit et., defaults to None"""
    )


def test_docstringdata__pti_ptc_ptf_ktb_ktlo_googleEpilogMultiline():
    epilog = adjust_epilog_for_test(
        """
        Et perferendis quia et sit maxime. Accusantium vel sint quam perspiciatis minus explicabo. Incidunt
        iste error autem impedit deserunt tempore quo aut odit.
        """
    )
    cmd = clig.Command(
        fun.pti_ptc_ptf_ktb_ktlo_googleEpilogMultiline,
        docstring_template=clig.GOOGLE_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "nesciunt beatae asperiores"
    assert data.epilog == epilog
    assert data.helps["a"] == clig._normalize_docstring(
        """Vel similique placeat. Nam enim perspiciatis qui earum voluptas quis. Perspiciatis ut
        vitae. Aspernatur ab ratione libero ex hic consequatur. Nam cupiditate earum. Nihil ea
        exercitationem ut."""
    )
    assert data.helps["b"] == clig._normalize_docstring(
        """Molestiae velit et expedita autem quam. Omnis dolorem placeat est. Quidem illum eveniet
        enim exercitationem aut qui dolore est et. Rerum est iste laudantium qui praesentium et. Et
        deserunt voluptates harum voluptas voluptates iste saepe consequatur."""
    )
    assert data.helps["c"] == clig._normalize_docstring(
        """Quae minima eligendi veniam aperiam libero temporibus quia qui atque. Velit ea aut vel
        quibusdam commodi id laboriosam inventore aliquam. Nam nisi itaque et sed dolor praesentium
        molestiae quisquam cupiditate. Voluptatem mollitia dolorem est deleniti repellat cum
        voluptatem voluptas sit.

        Sit animi dolore neque libero voluptatibus. Illum voluptatum ullam distinctio quisquam sequi
        delectus quia similique sit. Id enim vel eius iure rerum veritatis eos rem et. Nemo est
        assumenda aut et quo. Et soluta corrupti amet perferendis maxime.

        Placeat aut consequatur vel quo impedit doloribus et in libero. Voluptas ducimus suscipit.
        Assumenda alias est sed asperiores similique id consequuntur. Voluptas rerum placeat
        perferendis possimus ratione at. Ea ut aut id explicabo voluptas."""
    )
    assert data.helps["d"] == clig._normalize_docstring(
        """Error ut architecto fugit natus qui tempora vitae. A sed sequi reprehenderit
        quia autem voluptatem enim. Numquam cum minus cum eos est. Illo voluptas ducimus minus ipsam
        quae dolores quam quo. Quod qui sed incidunt rerum sed. Incidunt repellendus est est labore
        laudantium quia voluptas ipsum.. Defaults to True."""
    )
    assert data.helps["e"] == clig._normalize_docstring(
        """Explicabo tenetur beatae consequuntur atque aut omnis et. Eveniet
        ipsum repellat voluptatibus sit.

        Placeat eum veritatis praesentium voluptates quia beatae
        repellendus suscipit. Sint neque deserunt quis. Incidunt quibusdam voluptatem animi voluptas
        in. Voluptas dolor aut quisquam.. Defaults to None."""
    )


def test_docstringdata__ptc_kti_ktb_cligDocMutiline():
    epilog = adjust_epilog_for_test(
        """
        Qui quidem quo eligendi officia ea quod ab tempore esse. Sapiente quasi est sint. Molestias et
        laudantium quidem laudantium animi voluptate asperiores illum. Adipisci tempora nesciunt dolores
        tempore consequatur amet. Aut ipsa ex.
        """
    )
    cmd = clig.Command(
        fun.ptc_kti_ktb_cligDocMutiline,
        docstring_template=clig.CLIG_DOCSTRING_WITH_EPILOG,
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Reprehenderit unde commodi doloremque rerum ducimus quam accusantium."
    assert data.epilog == epilog
    assert data.helps["a"] == "Dicta et optio dicta."
    assert data.helps["b"] == "Dolorum voluptate voluptas nisi."
    assert data.helps["c"] == clig._normalize_docstring(
        """Asperiores quisquam odit voluptates et eos incidunt. Maiores minima provident doloremque aut
        dolorem. Minus natus ab voluptatum totam in. Natus consectetur modi similique rerum excepturi
        delectus aut."""
    )


def test_docstringdata__pn_pn_knb_kni_numpyEpilogNoType():
    cmd = clig.Command(
        fun.pn_pn_knb_kni_numpyEpilogNoType, docstring_template=clig.NUMPY_DOCSTRING_WITH_EPILOG_NOTYPES
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Odio est rerum iure porro rerum voluptatum libero magnam."
    assert data.epilog == "In vitae ut distinctio optio corrupti cumque rerum quasi aut."
    assert data.helps["a"] == "hic omnis sint"
    assert data.helps["b"] == "Ut rem quis delectus."
    assert data.helps["c"] == "Et tenetur modi ea., by default False"
    assert data.helps["d"] == "recusandae autem vero, by default 123"


def test_docstringdata__pn_pn_knb_kni_googleEpilogNoType():
    cmd = clig.Command(
        fun.pn_pn_knb_kni_googleEpilogNoType, docstring_template=clig.GOOGLE_DOCSTRING_WITH_EPILOG_NOTYPES
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Odio est rerum iure porro rerum voluptatum libero magnam."
    assert data.epilog == "In vitae ut distinctio optio corrupti cumque rerum quasi aut."
    assert data.helps["a"] == "Quasi veniam facere et."
    assert data.helps["b"] == clig._normalize_docstring(
        """Quis ex modi vel sed ea dolorum magnam. Ut veniam veniam minus. Laboriosam voluptatem quod et. Et
            eaque sint quasi libero mollitia."""
    )
    assert data.helps["c"] == "architecto non voluptas. Defaults to False."
    assert (
        data.helps["d"]
        == "Omnis laboriosam aut saepe nobis consequatur nihil eos accusantium.. Defaults to 123."
    )


def test_docstringdata__pn_pn_knb_kni_sphinxEpilogNoType():
    cmd = clig.Command(
        fun.pn_pn_knb_kni_sphinxEpilogNoType, docstring_template=clig.SPHINX_DOCSTRING_WITH_EPILOG_NOTYPES
    )
    data = cmd._collect_docstring_data_using_template()
    assert data is not None
    assert data.description == "Odio est rerum iure porro rerum voluptatum libero magnam."
    assert data.epilog == "In vitae ut distinctio optio corrupti cumque rerum quasi aut."
    assert data.helps["a"] == "aperiam enim voluptate"
    assert data.helps["b"] == "Totam voluptas porro est sint."
    assert data.helps["c"] == "Iusto impedit numquam ut., defaults to False"
    assert data.helps["d"] == "Corporis quis fugit eveniet rerum., defaults to 123"



# cSpell: disable
import os
import sys
import inspect

sys.path.insert(0, os.path.abspath(os.path.dirname(__file__) + "/../src"))
from clig import clig  # protected functions


def test_normalize_docstring():
    def foo():
        """A multi-line
        docstring.
        """

    def bar():
        """
        A multi-line
        docstring.
        """

    def aba(a, b):
        """My function

        My sumary

        Parameters
        ----------
        a : str
            one string
        b : ing
            one int
        """

    def gue(a, b):
        """
        My function

        My sumary

        Parameters
        ----------
        a : str
            one string
        b : ing
            one int

        """

    assert clig._normalize_docstring(foo.__doc__) == clig._normalize_docstring(bar.__doc__)
    assert clig._normalize_docstring(aba.__doc__) == clig._normalize_docstring(gue.__doc__)


def test_normalize_docstring_with_inspect():
    def foo(a: int, b: str, c: float, d: bool = True, e: list[str] | None = None) -> None:
        """Fugit voluptatibus enim odit velit facilis.

        Neque dolores expedita repellat in perspiciatis dolorem aliquid et. Commodi fugit minima
        laudantium beatae et ut. Id possimus soluta magnam quisquam laboriosam impedit.

        Ad quaerat ut culpa aut iure id quia. Ut aut alias adipisci quia. Veritatis ratione
        dignissimos laborum. Molestiae molestias id earum.

        Nesciunt quas corrupti tenetur officiis occaecati asperiores eaque. Qui voluptas ut ea dolor
        et harum beatae quos. Est tenetur ut ipsum. Eveniet rem beatae error eum voluptatem tempora
        velit in. Ea doloribus similique.

        Parameters
        ----------
        - `a` (`int`):
            Quidem natus sunt molestiae et reprehenderit voluptas optio.

        - `b` (`str`):
            Unde rerum aut a et assumenda fugit dolorem eligendi corrupti.

        - `c` (`float`):
            Dolorum officiis totam aspernatur fuga voluptas similique.

        - `d` (`bool`, optional): Defaults to `True`.
            Ducimus sunt eum in vel voluptatibus aut facere perspiciatis.

        - `e` (`list[str] | None`, optional): Defaults to `None`.
            Sit et consequatur a asperiores sequi sint dolores id ipsam.

        Returns
        -------
        `tuple[str, ...]`:
            illo odit ut
        """
        pass
        assert foo.__doc__ is not None
        assert clig._normalize_docstring(foo.__doc__) == inspect.cleandoc(foo.__doc__)


def test_normalize_docstring_with_one_line():
    def one_line():
        """Test of one line in the docsting"""

    def one_line_with_lines():
        """
        Test of one line in the docsting
        """

    assert clig._normalize_docstring(one_line.__doc__) == clig._normalize_docstring(
        one_line_with_lines.__doc__
    )


def test_normalize_docstring_multiline():
    def foo():
        """
        Fuga nemo provident vero odio qui sint et aut veritatis. Facere necessitatibus ut. Voluptatem
        natus natus veritatis earum. Reprehenderit voluptate dolorem dolores consequuntur magnam impedit
        eius. Est ut nisi aut accusamus.
        """
        pass

    assert (
        clig._normalize_docstring(foo.__doc__)
        == """Fuga nemo provident vero odio qui sint et aut veritatis. Facere necessitatibus ut. Voluptatem
natus natus veritatis earum. Reprehenderit voluptate dolorem dolores consequuntur magnam impedit
eius. Est ut nisi aut accusamus."""
    )



# cSpell: disable
import sys
from pathlib import Path

this_dir = Path(__file__).parent

sys.path.insert(0, str((this_dir).resolve()))
sys.path.insert(0, str((this_dir / "../src").resolve()))

from clig.clig import _normalize_docstring, Command, _ArgumentData, EMPTY
import functions as fun


def test_argumentdata_simple():
    cmd = Command(fun.pn_knc_noDoc)
    arg_1, arg_2 = cmd.argument_data
    assert arg_1 == _ArgumentData("first")
    assert arg_2 == _ArgumentData("second", typeannotation=str, default="test")


def test_argumentdata_with_types():
    cmd = Command(fun.pn_pt_kti_noDoc)
    arg_a, arg_b, arg_c = cmd.argument_data
    assert arg_a == _ArgumentData("a")
    assert arg_b == _ArgumentData("b", typeannotation=float)
    assert arg_c == _ArgumentData("c", typeannotation=int, default=123)


def test_argumentdata_with_types_and_helps():
    cmd = Command(fun.ptc_kti_ktb_cligDocMutiline)
    arg_a, arg_b, arg_c = cmd.argument_data
    assert arg_a == _ArgumentData("a", typeannotation=str, help="Dicta et optio dicta.")
    assert arg_b == _ArgumentData(
        "b", typeannotation=int, default=123, help="Dolorum voluptate voluptas nisi."
    )
    assert arg_c == _ArgumentData(
        "c",
        typeannotation=bool,
        default=True,
        help=_normalize_docstring(
            """Asperiores quisquam odit voluptates et eos incidunt. Maiores minima provident doloremque aut
            dolorem. Minus natus ab voluptatum totam in. Natus consectetur modi similique rerum excepturi
            delectus aut."""
        ),
    )


def test_argumentdata_with_types_and_metadata():
    cmd = Command(fun.ptcm_ptcm_ktb)
    arga, argb, argc = cmd.argument_data
    assert arga == _ArgumentData(
        name="a",
        default=EMPTY,
        typeannotation=str,
        help=None,
        make_flag=None,
        argument_group=None,
        mutually_exclusive_group=None,
        flags=["-f", "--first"],
        kwargs={"help": "The first argument"},
    )
    assert argb == _ArgumentData(
        "b", flags=[], typeannotation=int, kwargs={"action": "store_const", "const": 123}
    )
    assert argc == _ArgumentData(
        "c",
        flags=[],
        typeannotation=bool,
        default=True,
    )


def test_argumentdata_posWithType_posBoolWithType_cligDoc():
    cmd = Command(fun.ptc_ptb_cligEpilog)
    arg_name, arg_flag = cmd.argument_data
    assert arg_name == _ArgumentData("name", typeannotation=str, help="Sequi deserunt est quia qui.")
    assert arg_flag == _ArgumentData(
        "flag", typeannotation=bool, help="Labore eius et voluptatem quos et consequatur dolores."
    )
